import asyncio

from tqdm.asyncio import tqdm

from reachable import is_reachable, is_reachable_async
from reachable.client import AsyncClient


def test_serp():
    result = is_reachable("https://google.com")
    result2 = is_reachable(["https://google.com", "https://bing.com"])
    assert isinstance(result, dict)
    assert len(result2) == 2


def test_same_urls():
    result = is_reachable(["https://google.com", "https://google.com"])
    assert isinstance(result, list)
    assert len(result) == 1
    assert result[0]["status_code"] == 200


def test_same_urls_async():
    urls = ["https://google.com", "https://google.com"]
    result = asyncio.run(is_reachable_async(urls))
    assert isinstance(result, list)
    assert len(result) == 1
    assert result[0]["status_code"] == 200


def test_async():
    urls = ["https://google.com", "https://bing.com"]

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        # No loop already exists so we crete one
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    try:
        result = loop.run_until_complete(
            asyncio.gather(*[is_reachable_async(url) for url in urls])
        )
    finally:
        loop.close()

    assert len(result) == 2


def test_async2():
    urls = ["https://google.com", "https://bing.com"]

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        # No loop already exists so we crete one
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    async def wrapper():
        async with AsyncClient(include_host=True) as client:
            tasks = [is_reachable_async(url, client=client) for url in urls]
            return await tqdm.gather(*tasks)

    try:
        result = loop.run_until_complete(wrapper())
    finally:
        loop.close()

    assert len(result) == 2



import os.path as osp

from reachable.main import is_reachable, is_reachable_async


__all__ = ["is_reachable", "is_reachable_async"]

version_path = osp.join(osp.dirname(__file__), "VERSION.md")
if osp.exists(version_path):
    with open(version_path, "r") as f:
        __version__ = f.readline()



import asyncio
import logging
import re
import ssl
from typing import Any, Dict, Optional
from urllib.parse import urlparse, urlunparse

import httpx
import tldextract
from fake_useragent import UserAgent
from playwright.async_api import Error, TimeoutError, async_playwright
from typing_extensions import Self

ua: Any = UserAgent(browsers=["chrome"], os="windows", platforms="pc", min_version=120)


class Client:
    _type: str = "classic"

    def __init__(
        self,
        headers: Optional[Dict[str, str]] = None,
        include_host: bool = False,
        ssl_fallback_to_http: bool = False,
        ensure_protocol_url: bool = False,
    ) -> None:
        transport: httpx.HTTPTransport = httpx.HTTPTransport(retries=2)
        timeout: int = 10
        if headers is None:
            headers = {
                "User-Agent": ua.random,
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
                "Accept-Language": "en-US;q=0.7,en;q=0.3",
                "Accept-Encoding": "gzip, deflate, br, zstd",
                "DNT": "1",
                "Connection": "keep-alive",
                "TE": "trailers",
            }

        self.client: httpx.Client = httpx.Client(
            transport=transport,
            timeout=timeout,
            headers=headers,
            http2=True,
        )
        self.include_host: bool = include_host
        self.ssl_fallback_to_http: bool = ssl_fallback_to_http
        self.ensure_protocol_url: bool = ensure_protocol_url

    def request(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        include_host: bool = False,
        content: Any = None,
        ssl_fallback_to_http: bool = False,
    ):
        resp: Optional[httpx.Response] = None
        include_host = include_host | self.include_host
        ssl_fallback_to_http = ssl_fallback_to_http or self.ssl_fallback_to_http

        if include_host is True and headers is None:
            headers = {"Host": tldextract.extract(url).fqdn}
        elif include_host is True and headers is not None and "Host" not in headers:
            headers["Host"] = tldextract.extract(url).fqdn

        try:
            resp = self.client.request(method, url, headers=headers, content=content)
        except ssl.SSLError as e:
            if ssl_fallback_to_http is True:
                resp = self.client.request(
                    method,
                    url.lower().replace("https://", "http://"),
                    headers=headers,
                    content=content,
                )
            else:
                raise e
        except ssl.SSLWantReadError:
            # From https://github.com/encode/httpx/discussions/2941#discussioncomment-7574569
            # SSLWantReadError does itself not cause any problems. All it indicates is
            # that there isn't enough data in the local buffer for decrypting more
            # incoming data, so more needs to be read from the socket. And that's where
            # the timeout is coming from.
            # So we just retry
            pass
        return resp

    def get(
        self,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        include_host: bool = False,
        ssl_fallback_to_http: bool = False,
    ) -> Optional[httpx.Response]:
        return self.request(
            "get", url, headers, include_host, ssl_fallback_to_http=ssl_fallback_to_http
        )

    def post(
        self,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        content: Any = None,
        include_host: bool = False,
        ssl_fallback_to_http: bool = False,
    ) -> Optional[httpx.Response]:
        return self.request(
            "post",
            url,
            headers,
            include_host,
            content,
            ssl_fallback_to_http=ssl_fallback_to_http,
        )

    def head(
        self,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        include_host: bool = False,
        ssl_fallback_to_http: bool = False,
    ) -> Optional[httpx.Response]:
        return self.request(
            "head",
            url,
            headers,
            include_host,
            ssl_fallback_to_http=ssl_fallback_to_http,
        )

    def close(self) -> None:
        self.client.close()


class AsyncClient:
    _type: str = "classic"

    def __init__(
        self,
        headers: Optional[Dict[str, str]] = None,
        include_host: bool = False,
        ssl_fallback_to_http: bool = False,
        ensure_protocol_url: bool = False,
    ) -> None:
        self.transport: httpx.AsyncHTTPTransport = httpx.AsyncHTTPTransport(retries=2)
        self.timeout: int = 10
        if headers is None:
            self.headers = {
                "User-Agent": ua.random,
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
                "Accept-Language": "en-US;q=0.7,en;q=0.3",
                "Accept-Encoding": "gzip, deflate, br, zstd",
                "DNT": "1",
                "Connection": "keep-alive",
                "TE": "trailers",
            }
        else:
            self.headers = headers

        self.include_host: bool = include_host
        self.ssl_fallback_to_http: bool = ssl_fallback_to_http
        self.ensure_protocol_url: bool = ensure_protocol_url

    async def open(self) -> None:
        self.client: httpx.AsyncClient = httpx.AsyncClient(
            transport=self.transport,
            timeout=self.timeout,
            headers=self.headers,
            http2=True,
        )

    async def close(self) -> None:
        await self.client.aclose()

    async def __aenter__(self) -> Self:
        await self.open()
        return self

    # See link below for why more parameters than expected
    # https://docs.python.org/3/reference/datamodel.html#asynchronous-context-managers
    async def __aexit__(self, *args: Any) -> Self:
        await self.close()
        return self

    async def request(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        include_host: bool = False,
        content: Any = None,
        ssl_fallback_to_http: bool = False,
    ) -> Optional[httpx.Response]:
        resp: Optional[httpx.Response] = None
        include_host = include_host or self.include_host
        ssl_fallback_to_http = ssl_fallback_to_http or self.ssl_fallback_to_http

        if include_host is True and headers is None:
            # TLDExtract has better subdomain/domain separation
            # compared to urllib's urlparse
            headers = {"Host": tldextract.extract(url).fqdn}
        elif include_host is True and headers is not None and "Host" not in headers:
            headers["Host"] = tldextract.extract(url).fqdn

        if self.ensure_protocol_url is True:
            parsed_url = urlparse(url)

            if parsed_url.scheme != "http" or parsed_url.scheme != "https":
                url_replaced = parsed_url._replace(scheme="https")
                # Replace "///" by "//" in case URL is parsed as path and not netloc
                url = urlunparse(url_replaced).replace("https:///", "https://")

        try:
            resp = await self.client.request(
                method, url, headers=headers, content=content
            )
        except ssl.SSLError as e:
            if ssl_fallback_to_http is True:
                resp = await self.client.request(
                    method,
                    url.lower().replace("https://", "http://"),
                    headers=headers,
                    content=content,
                )
            else:
                raise e
        except ssl.SSLWantReadError:
            # From https://github.com/encode/httpx/discussions/2941#discussioncomment-7574569
            # SSLWantReadError does itself not cause any problems. All it indicates is
            # that there isn't enough data in the local buffer for decrypting more
            # incoming data, so more needs to be read from the socket. And that's where
            # the timeout is coming from.
            # So we just retry
            pass

        return resp

    async def get(
        self,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        include_host: bool = False,
        ssl_fallback_to_http: bool = False,
    ) -> Optional[httpx.Response]:
        return await self.request(
            "get", url, headers, include_host, ssl_fallback_to_http=ssl_fallback_to_http
        )

    async def post(
        self,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        content: Any = None,
        include_host: bool = False,
        ssl_fallback_to_http: bool = False,
    ) -> Optional[httpx.Response]:
        return await self.request(
            "post",
            url,
            headers,
            include_host,
            content,
            ssl_fallback_to_http=ssl_fallback_to_http,
        )

    async def head(
        self,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        include_host: bool = False,
        ssl_fallback_to_http: bool = False,
    ) -> Optional[httpx.Response]:
        return await self.request(
            "head",
            url,
            headers,
            include_host,
            ssl_fallback_to_http=ssl_fallback_to_http,
        )


class AsyncPlaywrightClient:
    _type: str = "browser"

    def __init__(
        self,
        headless: bool = False,
        ssl_fallback_to_http: bool = False,
        ensure_protocol_url: bool = False,
        executable_path: Optional[str] = None,
    ):
        self.playwright = None
        self.playwright_manager = async_playwright()

        self.browser = None

        self.ssl_fallback_to_http: bool = ssl_fallback_to_http
        self.ensure_protocol_url: bool = ensure_protocol_url
        self.headless: bool = headless
        self.executable_path: Optional[str] = executable_path

    async def open(self) -> None:
        self.playwright = await self.playwright_manager.__aenter__()
        self.browser = await self.playwright.chromium.launch(
            headless=self.headless, executable_path=self.executable_path
        )

    async def close(self) -> None:
        await self.browser.close()
        await self.playwright.stop()
        await self.playwright_manager.__aexit__()

    async def __aenter__(self) -> Self:
        await self.open()
        return self

    async def __aexit__(self, *args: Any) -> Self:
        await self.close()

    @staticmethod
    async def block_resources(route, request):
        # Manifest is the document for installing app on home screen on
        #   mobile defined by <link rel="manifest" href=""/>
        # TextTrack is for resources with videos
        if request.resource_type in [
            "image",
            "media",
            "stylesheet",
            "font",
            "manifest",
            "texttrack",
        ]:
            # Block the request
            await route.abort()
        else:
            # Continue with the request
            await route.continue_()

    @staticmethod
    async def _get_page_content(page, retries=3, delay=1):
        for i in range(retries):
            try:
                return await page.content()
            # Playwright is triggering `Error` exception or `UnboundLocalError`
            # which are not specific errors, so we catch them all
            except Exception:
                await asyncio.sleep(delay)

        raise Exception("Failed to get page content after multiple retries.")

    async def request(
        self,
        url: str,
        ssl_fallback_to_http: bool = False,
    ) -> Optional[httpx.Response]:
        ssl_fallback_to_http = ssl_fallback_to_http or self.ssl_fallback_to_http

        if self.ensure_protocol_url is True:
            parsed_url = urlparse(url)

            if parsed_url.scheme != "http" or parsed_url.scheme != "https":
                url_replaced = parsed_url._replace(scheme="https")
                # Replace "///" by "//" in case URL is parsed as path and not netloc
                url = urlunparse(url_replaced).replace("https:///", "https://")

        # We need a mutable data structure like a list or dictionary to hold the
        # response, as mutable structures can be modified inside nested functions.
        resp_obj = {"response": None, "final_url": url}

        page = await self.browser.new_page()

        # Register the route to block specific resources
        await page.route("**/*", AsyncPlaywrightClient.block_resources)

        async def response_handler(response):
            # If URL is "about:blank" it means this is the initial load,
            # so the first request must be the URL we are looking for
            # to handle correctly redirects.
            if page.url == "about:blank":
                resp_obj["response"] = response

        page.on("response", response_handler)

        content: str = ""
        try:
            await page.goto(url, timeout=60000)
            # Wait for all network requests in order to have the response object
            # and the HTML generated by an eventual React or Vue framework.
            await page.wait_for_load_state("networkidle")
            content = await AsyncPlaywrightClient._get_page_content(page, delay=2)
        except TimeoutError:
            raise httpx.ConnectTimeout("Connection timeout")
        except Error as e:
            # All Chronium errors are listed here:
            # https://source.chromium.org/chromium/chromium/src/+/main:net/base/net_error_list.h
            # DNS error, the domain has not been resolved
            if "NAME_NOT_RESOLVED" in str(e) or "CONNECTION_REFUSED" in str(e):
                raise httpx.ConnectError(
                    "[Errno 8] nodename nor servname provided, or not known"
                )
            elif "TIMED_OUT" in str(e):
                raise httpx.ConnectTimeout("Connection timeout")
            elif ("_SSL_" in str(e)) or ("_CERT_" in str(e)) and ssl_fallback_to_http:
                await page.close()
                await page.goto(url.replace("https://", "http://"), timeout=60000)
                # Wait for all network requests in order to have the response object
                # and the HTML generated by an eventual React or Vue framework.
                await page.wait_for_load_state("networkidle")
                content = await AsyncPlaywrightClient._get_page_content(page, delay=2)
            else:
                result = re.findall(r"net::([A-Z_0-9]*)", e.message)
                if len(result) > 0:
                    error = Error(result[0])
                    error.name = result[0]
                    raise error
                else:
                    raise e
        except Exception as e:
            raise e
        finally:
            await page.close()

        # Building the response
        resp: Optional[httpx.Response] = None
        if resp_obj["response"] is not None:
            headers = httpx.Headers(resp_obj["response"].headers)
            # When we build the Response, httpx will try to decompress the content
            # given the content-encoding value in the headers. Since it already
            # has been decompressed, we mark it as "identity" which
            # mean no compression
            headers["content-encoding"] = "identity"
            req = httpx.Request(method="get", url=resp_obj["response"].url)
            resp = httpx.Response(
                request=req,
                status_code=resp_obj["response"].status,
                headers=headers,
                content=content.encode(),
            )

        return resp

    async def get(
        self, url: str, ssl_fallback_to_http: bool = False
    ) -> Optional[httpx.Response]:
        return await self.request(url, ssl_fallback_to_http=ssl_fallback_to_http)

    async def post(
        self, url: str, ssl_fallback_to_http: bool = False
    ) -> Optional[httpx.Response]:
        logging.warning(
            "Using Playwright client, all requests are GET requests through the browser"
        )
        return await self.request(url, ssl_fallback_to_http=ssl_fallback_to_http)

    async def head(
        self, url: str, ssl_fallback_to_http: bool = False
    ) -> Optional[httpx.Response]:
        logging.warning(
            "Using Playwright client, all requests are GET requests through the browser"
        )
        return await self.request(url, ssl_fallback_to_http=ssl_fallback_to_http)



import asyncio
import hashlib
import os
import random
import ssl
import time
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse, urlunparse

import httpx
import tldextract
from tqdm import tqdm

from reachable.aio_client import AioAsyncClient
from reachable.client import AsyncClient, AsyncPlaywrightClient, Client


def is_reachable(
    url: Union[List[str], str],
    headers: Optional[Dict[str, str]] = None,
    include_host: bool = True,
    sleep_between_requests: bool = True,
    head_optim: bool = True,
    include_response: bool = False,
    client: Optional[Client] = None,
    ssl_fallback_to_http: bool = False,
    check_parking_domain: bool = False,
) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    return_as_list: bool = True
    url_list: List[str] = []

    if isinstance(url, str):
        url_list = [url]
        return_as_list = False
    elif isinstance(url, list):
        url_list = url
    else:
        raise ValueError(f"URL(s) of type {type(url)} is not supported")

    close_client: bool = True
    if client is None:
        client = Client(
            headers=headers,
            include_host=include_host,
            ssl_fallback_to_http=ssl_fallback_to_http,
        )
    else:
        close_client = False

    # Only keep unique URLs to avoid requesting same URL multiple times
    url_list = list(set(url_list))

    results: List[Dict[str, Any]] = []
    iterator: Union[List[str], tqdm] = url_list
    if return_as_list is True:
        iterator = tqdm(url_list)

    for elt in iterator:
        resp: Optional[httpx.Response] = None
        to_return: Dict[str, Any] = {
            "original_url": elt,
            "status_code": -1,
            "success": False,
            "error_name": None,
            "cloudflare_protection": False,
            "has_js_redirect": False,
        }

        resp, to_return["error_name"] = do_request(
            client,
            elt,
            head_optim=head_optim,
            sleep_between_requests=sleep_between_requests,
        )

        # Then we handle redirects
        if resp is not None and 400 > resp.status_code >= 300:
            to_return["error_name"] = None
            to_return["redirect"], resp, to_return["error_name"] = handle_redirect(
                client, resp
            )

            if to_return["redirect"]["final_url"] is not None:
                to_return["final_url"] = to_return["redirect"]["final_url"]

        if resp is not None:
            # Success
            if 300 > resp.status_code >= 200:
                to_return["success"] = True

            to_return["status_code"] = resp.status_code

            if b"cloudflareinsights.com" in resp.content:
                to_return["cloudflare_protection"] = True
            elif "cf-ray" in resp.headers:
                to_return["cloudflare_protection"] = True

            # Since really detecting JS redirects is not doable, we only detect
            # some cases and flag it has JS redirect. Of course it needs more tests
            # with some frameworks like selenium.
            if b"DOMContentLoaded" in resp.content and b"location.href" in resp.content:
                to_return["has_js_redirect"] = True

            if check_parking_domain is True:
                to_return["is_parking_domain"] = is_parking_domain(
                    client,
                    str(resp.url),
                    head_optim=head_optim,
                    sleep=sleep_between_requests,
                )

        if include_response is True:
            to_return["response"] = resp

        results.append(to_return)

    if close_client is True:
        client.close()

    if return_as_list is False:
        return results[0]
    else:
        return results


async def is_reachable_async(
    url: Union[List[str], str],
    headers: Optional[Dict[str, str]] = None,
    include_host: bool = True,
    sleep_between_requests: bool = True,
    head_optim: bool = True,
    include_response: bool = False,
    client: Optional[AsyncClient] = None,
    ssl_fallback_to_http: bool = False,
    check_parking_domain: bool = False,
) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    return_as_list: bool = True
    url_list: List[str] = []

    if isinstance(url, str):
        url_list = [url]
        return_as_list = False
    elif isinstance(url, list):
        url_list = url
    else:
        raise ValueError(f"URL(s) of type {type(url)} is not supported")

    close_client: bool = True
    if client is None:
        client = AsyncClient(
            headers=headers,
            include_host=include_host,
            ssl_fallback_to_http=ssl_fallback_to_http,
        )
        await client.open()
    else:
        close_client = False

    # Only keep unique URLs to avoid requesting same URL multiple times
    url_list = list(set(url_list))

    results: List[Dict[str, Any]] = []
    iterator: Union[List[str], tqdm] = url_list
    if return_as_list is True:
        iterator = tqdm(url_list)

    for elt in iterator:
        resp: Optional[httpx.Response] = None
        to_return: Dict[str, Any] = {
            "original_url": elt,
            "status_code": -1,
            "success": False,
            "error_name": None,
            "cloudflare_protection": False,
            "has_js_redirect": False,
        }

        # I don't know why but sometimes a TypeError is raised with the message
        # "an integer is required". This only happens when a httpx.ConnectError
        # has just been raised, tried different fixes without any success.
        # The problem appears to appear in the async process so the error
        # is not catchable here but where the async job has been called.
        # Looks like using `asyncio.create_task` fix the problem (thks ChatGPT).
        resp, to_return["error_name"] = await asyncio.create_task(
            do_request_async(
                client,
                elt,
                head_optim=head_optim,
                sleep_between_requests=sleep_between_requests,
            )
        )

        # If the request has been made by a browser client and the final URL doesn't
        # match the initial one, it has been redirected.
        # Redirects are handled transparently, so we need to populate `to_return`
        # with information that we have.
        if (
            client._type == "browser"
            and resp is not None
            and len(str(getattr(resp, "url", ""))) > 0
            and str(resp.url) != elt
            and 300 > resp.status_code >= 200
        ):
            to_return["redirect"] = {
                "chain": [str(resp.url)],
                "final_url": str(resp.url),
                "tld_match": is_tlds_matching(elt, str(resp.url), strict_suffix=False),
            }
            to_return["final_url"] = str(resp.url)

        # Then we handle redirects
        if resp is not None and 400 > resp.status_code >= 300:
            to_return["error_name"] = None
            (
                to_return["redirect"],
                resp,
                to_return["error_name"],
            ) = await handle_redirect_async(client, resp, head_optim=head_optim)

            if to_return["redirect"]["final_url"] is not None:
                to_return["final_url"] = to_return["redirect"]["final_url"]

        if resp is not None:
            # Success
            if 300 > resp.status_code >= 200:
                to_return["success"] = True

            to_return["status_code"] = resp.status_code

            if b"cloudflareinsights.com" in resp.content:
                to_return["cloudflare_protection"] = True
            elif "cf-ray" in resp.headers:
                to_return["cloudflare_protection"] = True

            # Since really detecting JS redirects is not doable, we only detect
            # some cases and flag it has JS redirect. Of course it needs more tests
            # with some frameworks like selenium.
            if b"DOMContentLoaded" in resp.content and b"location.href" in resp.content:
                to_return["has_js_redirect"] = True

            if check_parking_domain is True:
                to_return["is_parking_domain"] = await is_parking_domain_async(
                    client,
                    str(resp.url),
                    head_optim=head_optim,
                    sleep=sleep_between_requests,
                )

        if include_response is True:
            to_return["response"] = resp

        results.append(to_return)

    if close_client is True:
        await client.close()

    if return_as_list is False:
        return results[0]
    else:
        return results


async def is_reachable_aio_async(
    url: Union[List[str], str],
    headers: Optional[Dict[str, str]] = None,
    include_host: bool = True,
    sleep_between_requests: bool = True,
    head_optim: bool = True,
    include_response: bool = False,
    client: Optional[AioAsyncClient] = None,
) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    return_as_list: bool = True
    url_list: List[str] = []

    if isinstance(url, str):
        url_list = [url]
        return_as_list = False
    elif isinstance(url, list):
        url_list = url
    else:
        raise ValueError(f"URL(s) of type {type(url)} is not supported")

    close_client: bool = True
    if client is None:
        client = AioAsyncClient(headers=headers, include_host=include_host)
        await client.open()
    else:
        close_client = False

    # Only keep unique URLs to avoid requesting same URL multiple times
    url_list = list(set(url_list))

    results: List[Dict[str, Any]] = []
    iterator: Union[List[str], tqdm] = url
    if return_as_list is True:
        iterator = tqdm(url_list)

    for elt in iterator:
        resp = None
        to_return: Dict[str, Any] = {
            "original_url": elt,
            "status_code": -1,
            "success": False,
            "error_name": None,
            "cloudflare_protection": False,
        }

        resp, to_return["error_name"] = await do_request_async(
            client,
            elt,
            head_optim=head_optim,
            sleep_between_requests=sleep_between_requests,
        )

        # Then we handle redirects
        if resp is not None and 400 > resp.status >= 300:
            to_return["error_name"] = None
            (
                to_return["redirect"],
                resp,
                to_return["error_name"],
            ) = await handle_redirect_async(client, resp)

            if to_return["redirect"]["final_url"] is not None:
                to_return["final_url"] = to_return["redirect"]["final_url"]

        if resp is not None:
            # Success
            if 300 > resp.status >= 200:
                to_return["success"] = True

            to_return["status_code"] = resp.status

            if b"cloudflareinsights.com" in resp.content:
                to_return["cloudflare_protection"] = True

        if include_response is True:
            to_return["response"] = resp

        results.append(to_return)

    if close_client is True:
        await client.close()

    if return_as_list is False:
        return results[0]
    else:
        return results


def do_request(
    client: Client,
    url: str,
    head_optim: bool = True,
    sleep_between_requests: bool = True,
) -> Tuple[Optional[httpx.Response], Optional[str]]:
    error_name: Optional[str] = None
    resp: Optional[httpx.Response] = None

    # We first use HEAD to optimize requests
    try:
        if sleep_between_requests is True:
            time.sleep(random.SystemRandom().uniform(1, 2))

        # "Classic" client is httpx, AioHttp, etc.
        # Otherwise it is a "browser" like Playwright, etc
        if head_optim is True and client._type == "classic":
            resp = client.head(url)
        else:
            resp = client.get(url)
    except httpx.ConnectError:
        error_name = "ConnectionError"
    except httpx.ConnectTimeout:
        error_name = "ConnectTimeout"
    except httpx.ReadTimeout:
        error_name = "ReadTimeout"
    except httpx.RemoteProtocolError:
        error_name = "RemoteProtocolError"
    except ssl.SSLError:
        error_name = "SSLError"
    except httpx.ReadError:
        # Usually this is when the server block us
        error_name = "ReadError"
    except Exception as e:
        if getattr(e, "name", None) is not None:
            error_name = e.name
        else:
            error_name = type(e).__name__

    # Sometimes, the 40X and 50X errors are generated because of the use of HEAD request
    # If client's type is a browser, the error is definitive.
    if (
        head_optim is True
        and resp is not None
        and resp.status_code >= 400
        and client._type == "classic"
    ):
        # Reset error & response
        error_name = None
        resp = None

        try:
            if sleep_between_requests is True:
                time.sleep(random.SystemRandom().uniform(1, 2))
            resp = client.get(url)
        except httpx.ConnectError:
            error_name = "ConnectionError"
        except httpx.ConnectTimeout:
            error_name = "ConnectTimeout"
        except httpx.ReadTimeout:
            error_name = "ReadTimeout"
        except httpx.RemoteProtocolError:
            error_name = "RemoteProtocolError"
        except ssl.SSLError:
            error_name = "SSLError"
        except httpx.ReadError:
            # Usually this is when the server block us
            error_name = "ReadError"
        except Exception as e:
            if getattr(e, "name", None) is not None:
                error_name = e.name
            else:
                error_name = type(e).__name__

    return resp, error_name


async def do_request_async(
    client: Union[AsyncClient, AsyncPlaywrightClient],
    url: str,
    head_optim: bool = True,
    sleep_between_requests: bool = True,
    ssl_fallback_to_http: bool = False,
) -> Tuple[Optional[httpx.Response], Optional[str]]:
    error_name: Optional[str] = None
    resp: Optional[httpx.Response] = None

    # We first use HEAD to optimize requests
    try:
        if sleep_between_requests is True:
            await asyncio.sleep(random.SystemRandom().uniform(1, 2))

        # "Classic" client is httpx, AioHttp, etc.
        # Otherwise it is a "browser" like Playwright, etc
        if head_optim is True and client._type == "classic":
            resp = await client.head(url, ssl_fallback_to_http=ssl_fallback_to_http)
        else:
            resp = await client.get(url, ssl_fallback_to_http=ssl_fallback_to_http)
    except httpx.ConnectError:
        error_name = "ConnectionError"
    except httpx.ConnectTimeout:
        error_name = "ConnectTimeout"
    except httpx.ReadTimeout:
        error_name = "ReadTimeout"
    except httpx.WriteTimeout:
        error_name = "ReadTimeout"
    except httpx.RemoteProtocolError:
        error_name = "RemoteProtocolError"
    except ssl.SSLError:
        error_name = "SSLError"
    except httpx.ReadError:
        # Usually this is when the server block us
        error_name = "ReadError"
    except Exception as e:
        if getattr(e, "name", None) is not None:
            error_name = e.name
        else:
            error_name = type(e).__name__

    # Sometimes, the 40X and 50X errors are generated because of the use of HEAD request
    # If client's type is a browser, the error is definitive.
    if (
        head_optim is True
        and resp is not None
        and resp.status_code >= 400
        and client._type == "classic"
    ):
        # Reset error & response
        error_name = None
        resp = None

        try:
            if sleep_between_requests is True:
                await asyncio.sleep(random.SystemRandom().uniform(1, 2))
            resp = await client.get(url, ssl_fallback_to_http=ssl_fallback_to_http)
        except httpx.ConnectError:
            error_name = "ConnectionError"
        except httpx.ConnectTimeout:
            error_name = "ConnectTimeout"
        except httpx.WriteTimeout:
            error_name = "ReadTimeout"
        except httpx.ReadTimeout:
            error_name = "ReadTimeout"
        except httpx.RemoteProtocolError:
            error_name = "RemoteProtocolError"
        except httpx.HTTPStatusError as e:
            # For whatever reason HTTPStatusError is raised for non 20X status code
            # which is documented in HTTPX's documentation but this is not what happens
            # when using sync mode so we standardize behavior here.
            resp = e.response
        except ssl.SSLError:
            error_name = "SSLError"
        except httpx.ReadError:
            # Usually this is when the server block us
            error_name = "ReadError"
        except Exception as e:
            if getattr(e, "name", None) is not None:
                error_name = e.name
            else:
                error_name = type(e).__name__

    return resp, error_name


def is_tlds_matching(url1: str, url2: str, strict_suffix: bool = True) -> bool:
    is_matching: bool = False
    tld_orig: Any = tldextract.extract(url1)
    # In some cases they add the HTTPS port
    tld_redi: Any = tldextract.extract(url2)

    # If it's the same example.com
    if tld_orig.domain == tld_redi.domain and tld_orig.suffix == tld_redi.suffix:
        is_matching = True
    # If we have example.io and example.com
    elif tld_orig.domain == tld_redi.domain and strict_suffix is False:
        is_matching = True
    # Local redirection: /somepath
    elif tld_redi.domain == "":
        is_matching = True
    else:
        is_matching = False

    return is_matching


def _get_new_url(response: httpx.Response) -> str:
    url_origin: str = str(response.url)
    redirect_url: str = response.headers.get("location", "unknown")

    # If it is a local redirection
    new_url: str = ""
    tld_redirect: Any = tldextract.extract(redirect_url)
    if tld_redirect.domain == "":
        parsed_url = urlparse(url_origin)
        url_replaced = parsed_url._replace(path=redirect_url)
        new_url = urlunparse(url_replaced)
    else:
        new_url = redirect_url

    return new_url


def handle_redirect(
    client: Client,
    resp: httpx.Response,
    sleep_between_requests: bool = True,
    head_optim: bool = True,
) -> Tuple[Dict[str, Any], Optional[httpx.Response], Optional[str]]:
    error_name: Optional[str] = None
    new_resp: Optional[httpx.Response] = None
    chain: List[str] = []
    data: Dict[str, Any] = {
        "chain": [],
        "final_url": None,
        "tld_match": False,
    }

    new_url: str = _get_new_url(resp)

    new_resp, error_name, chain = follow_redirect(
        client,
        new_url,
        sleep_between_requests=sleep_between_requests,
        head_optim=head_optim,
    )

    data["chain"] = chain
    if new_resp is not None:
        data["final_url"] = str(new_resp.url)
        data["tld_match"] = is_tlds_matching(
            str(new_resp.url), data["final_url"], strict_suffix=False
        )

    return data, new_resp, error_name


async def handle_redirect_async(
    client: AsyncClient,
    resp: httpx.Response,
    sleep_between_requests: bool = True,
    head_optim: bool = True,
) -> Tuple[Dict[str, Any], Optional[httpx.Response], Optional[str]]:
    error_name: Optional[str] = None
    new_resp: Optional[httpx.Response] = None
    chain: List[str] = []
    data: Dict[str, Any] = {
        "chain": [],
        "final_url": None,
        "tld_match": False,
    }

    new_url: str = _get_new_url(resp)

    new_resp, error_name, chain = await follow_redirect_async(
        client,
        new_url,
        sleep_between_requests=sleep_between_requests,
        head_optim=head_optim,
    )

    data["chain"] = chain
    if new_resp is not None:
        data["final_url"] = str(new_resp.url)
        data["tld_match"] = is_tlds_matching(
            str(new_resp.url), data["final_url"], strict_suffix=False
        )

    return data, new_resp, error_name


def follow_redirect(
    client: Client,
    url: str,
    depth: int = 5,
    sleep_between_requests: bool = True,
    head_optim: bool = True,
) -> Tuple[Optional[httpx.Response], Optional[str], List[str]]:
    if depth <= 0:
        return None, "Max depth reached", []

    chain: List[str] = [url]
    resp, error_name = do_request(
        client,
        url,
        head_optim=head_optim,
        sleep_between_requests=sleep_between_requests,
    )

    # Has redirect
    if resp is not None and 400 > resp.status_code >= 300:
        new_url: str = _get_new_url(resp)
        nresp, error_name, tchain = follow_redirect(
            client,
            new_url,
            depth=depth - 1,
            sleep_between_requests=sleep_between_requests,
            head_optim=head_optim,
        )
        chain += tchain
        return nresp, error_name, chain
    else:
        return resp, error_name, chain


async def follow_redirect_async(
    client: AsyncClient,
    url: str,
    depth: int = 5,
    sleep_between_requests: bool = True,
    head_optim: bool = True,
) -> Tuple[Optional[httpx.Response], Optional[str], List[str]]:
    if depth <= 0:
        return None, "Max depth reached", []

    chain: List[str] = [url]
    resp, error_name = await do_request_async(
        client,
        url,
        head_optim=head_optim,
        sleep_between_requests=sleep_between_requests,
    )

    # Has redirect
    if resp is not None and 400 > resp.status_code >= 300:
        new_url: str = _get_new_url(resp)
        nresp, error_name, tchain = await follow_redirect_async(
            client,
            new_url,
            depth=depth - 1,
            sleep_between_requests=sleep_between_requests,
            head_optim=head_optim,
        )
        chain += tchain
        return nresp, error_name, chain
    else:
        return resp, error_name, chain


def _replace_url_path(url: str, path: str) -> str:
    parsed_url = urlparse(url)
    url_replaced = parsed_url._replace(query="", path=path)
    return urlunparse(url_replaced)


async def is_parking_domain_async(
    client: AsyncClient, url: str, head_optim: bool = True, sleep: bool = False
) -> bool:
    # Set random URL and if it returns 200, it is a parked domain since
    # they always answer with 200 or redirect
    rand = hashlib.sha512(os.urandom(128)).hexdigest()
    new_url = _replace_url_path(url, path=f"{rand[:64]}/{rand[65:]}")
    result, _ = await do_request_async(
        client, new_url, head_optim=head_optim, sleep_between_requests=sleep
    )
    return result.status_code < 400


def is_parking_domain(
    client: AsyncClient, url: str, head_optim: bool = True, sleep: bool = False
) -> bool:
    # Set random URL and if it returns 200, it is a parked domain since
    # they always answer with 200 or redirect
    rand = hashlib.sha512(os.urandom(128)).hexdigest()
    new_url = _replace_url_path(url, path=f"{rand[:64]}/{rand[65:]}")
    result, _ = do_request(
        client, new_url, head_optim=head_optim, sleep_between_requests=sleep
    )
    return result.status_code < 400



import asyncio
from types import TracebackType
from typing import Any, Awaitable, List, Optional, Type

from tqdm.asyncio import tqdm


# Based on
# https://medium.com/@cgarciae/making-an-infinite-number-of-requests-with-python-aiohttp-pypeln-3a552b97dc95
class TaskPool(object):
    def __init__(self, workers: int, use_tqdm: bool = True):
        self._semaphore: asyncio.Semaphore = asyncio.Semaphore(workers)
        self._tasks: List[Awaitable[Any]] = []
        self._results: List[Any] = []
        self.use_tqdm: bool = use_tqdm

    async def put(self, coro: Awaitable[Any]) -> None:
        await self._semaphore.acquire()

        task: Any = asyncio.ensure_future(coro)
        task.add_done_callback(self._on_task_done)
        self._tasks.append(task)

    def _on_task_done(self, task: Any) -> None:
        try:
            res = task.result()
            self._results.append(res)
        except Exception as e:
            print(f"Task raised an exception: {e}")
        finally:
            self._semaphore.release()

    async def join(self) -> None:
        if self.use_tqdm:
            await tqdm.gather(*self._tasks)
        else:
            await asyncio.gather(*self._tasks)

    async def __aenter__(self) -> Any:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.join()



import asyncio

from tqdm.asyncio import tqdm

from reachable import is_reachable_async
from reachable.client import AsyncPlaywrightClient
from reachable.pool import TaskPool


# Please provide your Chromium path here
chromium_path: str = ""


def test_async():
    urls = ["https://google.com", "https://bing.com"]

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        # No loop already exists so we crete one
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    async def wrapper():
        async with AsyncPlaywrightClient(
            headless=True,
            executable_path=chromium_path,
        ) as client:
            tasks = [client.request(url) for url in urls]
            return await tqdm.gather(*tasks)

    try:
        result = loop.run_until_complete(wrapper())
    finally:
        loop.close()

    assert len(result) == len(urls)


async def _executor(url, client):
    result = await is_reachable_async(
        url,
        client=client,
        include_response=True,
    )
    return result


async def _wrapper(urls):
    async with AsyncPlaywrightClient(
        headless=True,
        executable_path=chromium_path,
        ensure_protocol_url=True,
        ssl_fallback_to_http=True,
    ) as client:
        tasks = TaskPool(workers=10)

        for url in tqdm(urls, desc="Loading data"):
            await tasks.put(_executor(url, client))

        await tasks.join()

    return tasks._results


def test_pool():
    urls = ["https://google.com", "https://bing.com"]

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        # No loop already exists so we crete one
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    results = loop.run_until_complete(_wrapper(urls))

    assert len(results) == len(urls)
